const core = @import("core");
const sys = @import("./sys.zig");

///
/// Metadata of an Oar archive file entry.
///
const Entry = extern struct {
    signature: [signature_magic.len]u8 = signature_magic,
    path: sys.Path = sys.Path.empty,
    data_offset: u64 = 0,
    data_length: u64 = 0,
    padding: [232]u8 = [_]u8{0} ** 232,

    comptime {
        const entry_size = @sizeOf(@This());

        if (entry_size != 512) @compileError("EntryBlock is not 512 bytes");
    }
};

///
/// [FindError.ArchiveUnsupported] occurs when trying to read a file that does not follow an Oar
/// archive format considered valid by this implemenatation.
///
/// [FindError.EntryNotFound] occurs when the queried entry was not found in the archive file.
///
pub const FindError = core.io.AccessError || error {
    ArchiveUnsupported,
    EntryNotFound,
};

///
/// Header data that every Oar archive file starts with at byte offset `0`.
///
const Header = extern struct {
    signature: [signature_magic.len]u8 = signature_magic,
    revision: u8 = revision_magic,
    entry_count: u32 = 0,
    padding: [502]u8 = [_]u8{0} ** 502,

    comptime {
        const size = @sizeOf(@This());

        if (size != 512) @compileError("Header is not 512 bytes");
    }
};

///
/// Attempts to find an [Entry] with a path name matching `path` in `archive_reader`.
///
/// An [Entry] value is returned if a match was found, otherwise [FindError] if it failed.
///
pub fn findEntry(archive_reader: sys.FileReader, path: sys.Path) FindError!Entry {
    var header = Header{};
    const header_size = @sizeOf(Header);
    const io = core.io;

    if ((try archive_reader.read(io.bytesOf(&header))) != header_size)
        return error.ArchiveUnsupported;

    if (!io.equals(u8, &header.signature, &signature_magic))
        return error.ArchiveUnsupported;

    if (header.revision != revision_magic) return error.ArchiveUnsupported;

    // Read file table.
    var head: u64 = 0;
    var tail: u64 = (header.entry_count - 1);
    const entry_size = @sizeOf(Entry);

    while (head <= tail) {
        var entry = Entry{};
        const midpoint = head + ((tail - head) / 2);
        const offset = header_size + (entry_size * midpoint);

        try archive_reader.seek(offset);

        if ((try archive_reader.read(io.bytesOf(&entry))) != entry_size)
            return error.ArchiveUnsupported;

        const comparison = path.compare(entry.path);

        if (comparison == 0) return entry;

        if (comparison > 0) {
            head = (midpoint + 1);
        } else {
            tail = (midpoint - 1);
        }
    }

    return error.EntryNotFound;
}

///
/// Magic revision number that this Oar software implementation understands.
///
const revision_magic = 0;

///
/// Magic identifier used to validate [Header] and [Block] data.
///
const signature_magic = [3]u8{'o', 'a', 'r'};
